# Copyright {{ year }} {{ author }}
# See LICENSE file for licensing details.
name: {{ name }}
description: |
  TODO: fill out the charm's description
summary: |
  TODO: fill out the charm's summary
series: {{ series }}

