# Copyright {{ year }} {{ author }}
# See LICENSE file for licensing details.
#
# This is only an example, and you should edit to suit your needs.
# If you don't need config, you can remove the file entirely.
options:
  thing:
    default: 🎁
    description: A thing used by the charm.
    type: string
